#include <iostream>
#include <fstream>
#include <cmath>
#include <vector>
#include <set>
#include <map>
#include <algorithm>
#include <stdio.h>

#include "TPostScript.h"
#include "TCanvas.h"
#include "TStyle.h"
#include "TGraphErrors.h"
#include "TGraph2D.h"
#include "TFile.h"
#include "TH1I.h"

/*  So, the point of this function is to read in a file of fill, eta, peak, and
 *  error values and use that information to construct a graph of the percent
 *  deviations over time for every eta bin.
 */

//  We need a struct here to keep track of all the information that we're
//  reading in from our input file.  We have to use that information in several
//  places, and this seems like the best way to organize that information.
struct DataPoint
{
    int fill;
    int eta;
    int events;
    double peak;
    double error;
    double mean;
    double meanError;
    //	This is the deviation from the mean of the graph for this point.
    double deviation;
    //	This is the ratio of the devation from the mean for this point to the
    //	mean (i.e. the percent deviation).
    double ratioDeviation;
    //	This is the weighted deviation for the point.
    double weightedDeviation;
    double chi2;
    int ndf;
    double chi2ndf;
};

//  Here are the functions that we're using in this file.
void printPointsTable( std::ostream &, DataPoint **, DataPoint **, int );
void printPointsHeader( std::ostream & );
void printPointRow( std::ostream &, DataPoint *, DataPoint * );
void printBadPointsTable(
    std::ostream &, std::set<DataPoint *>, std::set<DataPoint *>
);
void printBadPointsHeader( std::ostream & );
void printBadPointRow( std::ostream &, DataPoint * );

//  We're going to assume that there is only one file that we're dealing with
//  here.  This whole "working with multiple sets of histograms" thing seems
//  like overkill at this point.
int main()
{
    const char* file="mip2006.gains.corrected";
    const char* postscript="percentDeviationOverTime.ps";

    //	There is some interest in saving these graphs for interactive use.
    const char* rootOutput="percentDeviationOverTime.root";

    //	This is the number of eta bins that we are expecting.
    const int neta = 40;

    //	This is for record keeping in the output files.
    std::cout << "input file:  " << file << std::endl;
    std::cout << "output file: " << postscript << std::endl;

    TPostScript *ps = new TPostScript( postscript );
    TFile outputFile( rootOutput, "RECREATE" );
    TCanvas *myCanvas = new TCanvas( "c", "", 100, 100, 600, 800 );
    myCanvas->SetBatch(1);
    int pad = 1;

    //	This is our input file.
    std::ifstream infile( file );

    //	We're going to be reading in all the data from the file into a reusable
    //	data structure (or set of data structures).  I think that, basically,
    //	we want an array of vectors of data points.  That is, we want, one
    //	vector per eta bin, and each vector will be filled with one data point
    //	per fill for that eta bin.
    std::vector<DataPoint> bins[neta];
    DataPoint curPoint;

    //	We also going to be keeping track of very low statistic bin-fills and
    //	of bin-fills that are known bad fits.
    std::set<DataPoint *> lowStatBinFills;
    std::set<DataPoint *> badBinFills;

    //	We also need to keep track of a set of fills.  Without this, we won't
    //	know whether the individual vectors are missing fills or not.  This set
    //	will be used to determine the "time" axis of the graph that we are
    //	generating.  You may be wondering why this set is a map.  That is
    //	because the fills will be mapped to their particular ordering time
    //	number using this data structure, allowing us to ignore the ordering of
    //	the eta vectors.
    std::map<int, int> fills;

    //	We need to keep track of the maximum number of events here.
    int maxEvents = 0;

    //	We need to calculate the min and max Chi^2/DOF values as well.
    double minchi2ndf = 0;
    double maxchi2ndf = 0;

    //	While we have stuff in the file. . . 
    while (
	infile	>> curPoint.fill
		>> curPoint.eta
		>> curPoint.peak
		>> curPoint.error
		>> curPoint.mean
		>> curPoint.meanError
		>> curPoint.events
		>> curPoint.chi2
		>> curPoint.ndf
    )
    {
	//  Calculating this here for efficiency.
	if ( curPoint.ndf != 0 )
	{
	    curPoint.chi2ndf = curPoint.chi2/curPoint.ndf;

	    if ( curPoint.chi2ndf > maxchi2ndf )
	    {
		maxchi2ndf = curPoint.chi2ndf;
	    }
	    if ( curPoint.chi2ndf < minchi2ndf )
	    {
		minchi2ndf = curPoint.chi2ndf;
	    }
	}

	bins[ curPoint.eta - 1 ].push_back(curPoint);
	//  This next line should create the map element if it does not exist,
	//  and it should have no negative side effects if the element does
	//  exist, except for causing some slow-down, most of which would occur
	//  regardless of how this was done.
	fills[curPoint.fill];
	//curPoint = new DataPoint();

	if ( curPoint.events > maxEvents )
	{
	    maxEvents = curPoint.events;
	}

	//  I'm worried, probablyl needlessly, about whether or not the error
	//  is ever less than zero.  It shouldn't be, but I'm going to check.
	if ( curPoint.error < 0 )
	{
	    fprintf(
		stderr, 
		"Bin %i at fill %i had an error less than zero: %f", 
		curPoint.eta,
		curPoint.fill,
		curPoint.error
	    );
	}
    }

    infile.close();

    //	Now we need to ensure that the ordering of the fills is mapped to the
    //	fill number in the fills map.
    std::map<int, int>::iterator myIter;
    int fillNumber = 0;

    //	This is primarily here because I don't trust libraries implemented by
    //	other people to follow standards.  I also don't trust libraries I've
    //	built myself to follow standards, so this really should not insult
    //	anybody.
    int oldFill = -1;

    for (
	myIter=fills.begin();
	myIter != fills.end();
	++fillNumber, ++myIter
    )
    {
	if (myIter->first < oldFill)
	{
	    std::cerr << "Iterating through the map did not function as advertised" << std::endl;
	    return 1;
	}

	oldFill = myIter->first;
	myIter->second = fillNumber;
    }

    char graphName[200];

    TGraphErrors *myGraph[neta];
    int numPoints;

    //	So, next we create the graphs that we're using to find our fits.  This
    //	should be relatively straight-forward now.
    for ( int ii = 0; ii < neta; ++ii )
    {
	myGraph[ ii ] = new TGraphErrors();
	//  Notice that we're setting the number of points in the graph first.
	myGraph[ ii ]->Set( bins[ii].size() );

	numPoints = bins[ii].size();

	for ( int jj = 0; jj < numPoints; ++jj )
	{
	    myGraph[ ii ]->SetPoint(
		jj, (double) fills[ bins[ii][jj].fill ]+1, bins[ii][jj].peak
	    );
	    myGraph[ ii ]->SetPointError( jj, 0, bins[ii][jj].error );

	    //	We're still worried(?) about peak that are equal to zero, so:
	    if ( bins[ii][jj].peak == 0 )
	    {
		std::cerr << "Got peak == 0 on bin " << ii << ", fill ";
		std::cerr << bins[ii][jj].fill << std::endl;
	    }
	}
    }

    //	Now that we have our graphs, we need to fit them.  We expect there to
    //	be a constant-over-time relationship between the points that we are
    //	plotting, so we'll use a constant, linear fit.  We'll need a bunch of
    //	these fitters.
    TF1 *fitters[neta];
    char fitname[200];

    for ( int ii = 0; ii < neta; ++ii )
    {
	sprintf( fitname, "Fit for eta %i", ii+1 );
	//  We're going from 1 to fills size plus 1 here because that should
	//  be the range of the graph (see the graph generation algorithm).
	fitters[ ii ] = new TF1( fitname, "pol0", 1, fills.size()+1 );

	myGraph[ ii ]->Fit( fitters[ ii ], "rq" );

	if ( ii % 2 == 0 )
	{
	    myCanvas->Update();
	    ps->NewPage();
	    myCanvas->Clear();
	    myCanvas->Divide( 1, 2 );
	    pad = 1;
	}

	myCanvas->cd( pad );

	//  We'll draw these, as a double check.
	sprintf( graphName, "Graph of eta %i", ii+1 );
        myGraph[ ii ]->SetTitle( graphName );
        myGraph[ ii ]->GetXaxis()->SetTitle( "\"Fill Time\"" );
        myGraph[ ii ]->GetYaxis()->SetTitle( "mip Position" );
        myGraph[ ii ]->GetYaxis()->SetRangeUser( 12, 25 );
        myGraph[ ii ]->GetXaxis()->SetRangeUser( 0, fills.size()+1 );
        myGraph[ ii ]->SetMarkerStyle( 25 );
        myGraph[ ii ]->SetMarkerSize( .5 );
        myGraph[ ii ]->SetMarkerColor( 4 );
    
        fitters[ ii ]->SetTitle( graphName );
        fitters[ ii ]->GetYaxis()->SetRangeUser( 12, 25 );
        fitters[ ii ]->GetXaxis()->SetRangeUser( 0, fills.size()+1 );
        fitters[ ii ]->SetLineWidth( 2 );
        fitters[ ii ]->SetLineColor( 2 );
        fitters[ ii ]->SetLineStyle( 1 );

        //  Draw the graph.
        myGraph[ ii ]->Draw( "AP" );
        fitters[ ii ]->Draw( "LSAME" );

	myGraph[ ii ]->Write();
	fitters[ ii ]->Write();

	++pad;
    }

    //	Next, we have to construct the 2D graph in time, eta bin, and percent
    //	deviation.  In order to do this, we need to calculate the percent
    //	deviation from the mean, as determined by the previous fit.
    double mean;

    //	Here we're also trying to find the mamximum and minimum weighted
    //	deviations for each eta bin, anlong with the corresponding percent
    //	errors for those bins.  Of course, all we really need is the individual
    //	data point, since that conveniently stores all that information.
    DataPoint * minErrors[ neta ];
    DataPoint * maxErrors[ neta ];

    //	Okay, as we're reading these in, we want to keep track of two different
    //	histograms.  The first one is going to be a histogram of events per
    //	eta/fill.  The second one is going to a histogram of events per
    //	eta/fill where peak < 10.
    TH1I *eventsHisto = 
	new TH1I( "eventsHisto", "eventsHisto", 200, 0, maxEvents+1 );
    TH1I *badEventsHisto = 
	new TH1I( "badEventsHisto", "badEventsHisto", 200, 0, 200 );

    //	We're also going to histogram chi^2 values, both for all fill/eta
    //	pairs, and for known bad fill/eta pairs.
    TH1I *chi2Histo = 
	new TH1I( "chi2Histo", "chi2Histo", 500, minchi2ndf-1, maxchi2ndf+2 );
    TH1I *badChi2Histo = 
	new TH1I(
	    "badChi2Histo", "badChi2Histo", 500, minchi2ndf-1, maxchi2ndf+2
	);

    for ( int ii = 0; ii < neta; ++ii )
    {
	minErrors[ ii ] = 0;
	maxErrors[ ii ] = 0;

	mean = fitters[ii]->GetParameter(0);
	//  This is just in case.  Should never happen.
	if ( mean == 0 )
	{
	    std::cerr << "Got mean = 0 for bin " << ii << ", skipping" << std::endl;
	    continue;
	}

	numPoints = bins[ii].size();

	for ( int jj = 0; jj < numPoints; ++jj )
	{
	    eventsHisto->Fill( bins[ii][jj].events );
	    if ( bins[ii][jj].ndf != 0 )
	    {
		chi2Histo->Fill( bins[ii][jj].chi2ndf );
	    }
	    else
	    {
		chi2Histo->Fill( maxchi2ndf+1 );
	    }

	    if ( bins[ii][jj].peak < 0 )
	    {
		badEventsHisto->Fill( bins[ii][jj].events );
		if ( bins[ii][jj].ndf != 0 )
		{
		    badChi2Histo->Fill( bins[ii][jj].chi2ndf );
		}
		else
		{
		    badChi2Histo->Fill( maxchi2ndf+1 );
		}
	    }

	    bins[ii][jj].deviation = bins[ii][jj].peak - mean;
	    //std::cout << bins[ii][jj].deviation << std::endl;;
	    bins[ii][jj].ratioDeviation = bins[ii][jj].deviation/mean;
	    //std::cout << bins[ii][jj].ratioDeviation << std::endl;;
	    bins[ii][jj].weightedDeviation =
		bins[ii][jj].deviation/bins[ii][jj].error;
	    //std::cout << bins[ii][jj].weightedDeviation << std::endl;;

	    //	We are not accumulating an error if either the number of
	    //	degrees of freedom is less than or equal to zero, or if the
	    //	peak is less than or equal to zero.  In both of these cases,
	    //	we *know* that the fit did not match any physics.  This will
	    //	miss bad fits, but it should at least eliminate these fits that
	    //	we know for certain are bad.
	    if ( bins[ii][jj].ndf > 0 && bins[ii][jj].peak > 0 )
	    {
		if ( minErrors[ ii ] == 0 
		     || minErrors[ ii ]->weightedDeviation
			    > bins[ii][jj].weightedDeviation
		)
		{
		    minErrors[ ii ] = &(bins[ii][jj]);
		}
		if ( maxErrors[ ii ] == 0 
		     || maxErrors[ ii ]->weightedDeviation
			    < bins[ii][jj].weightedDeviation
		)
		{
		    maxErrors[ ii ] = &(bins[ii][jj]);
		}
	    }
	    else if ( bins[ii][jj].ndf <= 0 )
	    {
		lowStatBinFills.insert( &bins[ii][jj] );
	    }
	    else
	    {
		badBinFills.insert( &bins[ii][jj] );
	    }
	}
    }

    //	Now, let's output a table of the min/max error points.
    std::ofstream blahStream( "blah.html" );
    printPointsTable( blahStream, minErrors, maxErrors, neta );
    printBadPointsTable( blahStream, lowStatBinFills, badBinFills );
    blahStream.close();

    //	Now we can fill in graph data.  We'll need an independent counter for
    //	the SetPoint function.
    int pointCounter = 0;

    //	And we'll a couple of actual graphs.  We can, however, determine
    //	apriori the maximum number of points that we will be storing in this
    //	graph.  The first graph is just a plot of the straight percent
    //	deviation for each poinnt.  The second graph is a plot of deviations
    //	weighted by the error on the individual point.
    TGraph2D *deviationGraph = new TGraph2D( fills.size() * neta );
    deviationGraph->SetName( "percent2DGraph" );
    TGraph2D *weightedDeviationGraph = new TGraph2D( fills.size() * neta );
    weightedDeviationGraph->SetName( "weighted2DGraph" );

    //	We're also going to create two additional graphs that are the same,
    //	except that they will not have the points that we know are bad.
    TGraph2D *deviationGraphSansBad = new TGraph2D( fills.size() * neta );
    deviationGraphSansBad->SetName( "percent2DGraphSansBad" );
    TGraph2D *weightedDeviationGraphSansBad =
	new TGraph2D( fills.size() * neta );
    weightedDeviationGraphSansBad->SetName( "weighted2DGraphSansBad" );

    for ( int ii = 0; ii < neta; ++ii )
    {
	numPoints = bins[ii].size();

	for ( int jj = 0; jj < numPoints; ++jj )
	{
	    deviationGraph->SetPoint(
		pointCounter,
		(double) fills[ bins[ii][jj].fill ] + 1,
		(double) ii + 1,
		bins[ii][jj].ratioDeviation
	    );
	    weightedDeviationGraph->SetPoint(
		pointCounter,
		(double) fills[ bins[ii][jj].fill ] + 1,
		(double) ii + 1,
		bins[ii][jj].weightedDeviation
	    );
	    if (
		!lowStatBinFills.count( &bins[ii][jj] )
		&& !badBinFills.count( &bins[ii][jj] )
	    )
	    {
		deviationGraphSansBad->SetPoint(
		    pointCounter,
		    (double) fills[ bins[ii][jj].fill ] + 1,
		    (double) ii + 1,
		    bins[ii][jj].ratioDeviation
		);
		weightedDeviationGraphSansBad->SetPoint(
		    pointCounter,
		    (double) fills[ bins[ii][jj].fill ] + 1,
		    (double) ii + 1,
		    bins[ii][jj].weightedDeviation
		);
	    }
	    ++pointCounter;
	    //fprintf( stderr, "Fill: %f, Eta: %f, Ratio: %f\n", (float) fills[ bins[ii][jj].fill ] +1, (float) ii + 1, (float) bins[ii][jj].ratioDeviation);
	}
    }

    myCanvas->Update();
    ps->NewPage();
    myCanvas->Clear();
    myCanvas->Divide( 1, 1 );
    myCanvas->cd( 1 );

    sprintf ( graphName, "Ratio of deviation-to-mean over time/eta bin" );

    //  Beautifing the graph.
    deviationGraph->SetTitle( graphName );

    //	This part of beautifying stuff takes a long time, so we'll warn.
    std::cout 
	<< "Generating graph.  This could take some time. . ." << std::endl;
    deviationGraph->GetXaxis()->SetTitle( "\"Fill Time\"" );
    deviationGraph->GetYaxis()->SetTitle( "Bin number" );
    deviationGraph->GetZaxis()->SetTitle( "Peak Dev. from Mean" );
    deviationGraph->GetXaxis()->SetRangeUser(
	0,  fills.size()+1
    );
    //deviationGraph->GetZaxis()->SetRangeUser( -1, 1 );
    deviationGraph->GetXaxis()->SetRangeUser(
	0,  neta+1
    );
    fprintf (stderr, "Got here\n" );

    //	Draw the graph.
    gStyle->SetPalette(1);
    deviationGraph->Draw("surf1");

    deviationGraph->Write();

    myCanvas->Update();
    ps->NewPage();
    myCanvas->Clear();
    myCanvas->Divide( 1, 1 );
    myCanvas->cd( 1 );

    sprintf ( graphName, "Weighted deviation over time/eta bin" );

    //  Beautifing the graph.
    weightedDeviationGraph->SetTitle( graphName );

    //	This part of beautifying stuff takes a long time, so we'll warn.
    std::cout 
	<< "Generating graph.  This could take some time. . ." << std::endl;
    weightedDeviationGraph->GetXaxis()->SetTitle( "\"Fill Time\"" );
    weightedDeviationGraph->GetYaxis()->SetTitle( "Bin number" );
    weightedDeviationGraph->GetZaxis()->SetTitle( "Peak Dev. from Mean" );
    weightedDeviationGraph->GetXaxis()->SetRangeUser(
	0,  fills.size()+1
    );
    //weightedDeviationGraph->GetZaxis()->SetRangeUser( -1, 1 );
    weightedDeviationGraph->GetXaxis()->SetRangeUser(
	0,  neta+1
    );
    fprintf (stderr, "Got here\n" );

    //	Draw the graph.
    weightedDeviationGraph->Draw("surf1");

    weightedDeviationGraph->Write();

    myCanvas->Update();
    ps->NewPage();
    myCanvas->Clear();
    myCanvas->Divide( 1, 1 );
    myCanvas->cd( 1 );

    sprintf (
	graphName,
	"Corrected ratio of deviation-to-mean over time/eta bin"
    );

    //  Beautifing the graph.
    deviationGraphSansBad->SetTitle( graphName );

    //	This part of beautifying stuff takes a long time, so we'll warn.
    std::cout 
	<< "Generating graph.  This could take some time. . ." << std::endl;
    deviationGraphSansBad->GetXaxis()->SetTitle( "\"Fill Time\"" );
    deviationGraphSansBad->GetYaxis()->SetTitle( "Bin number" );
    deviationGraphSansBad->GetZaxis()->SetTitle( "Peak Dev. from Mean" );
    deviationGraphSansBad->GetXaxis()->SetRangeUser(
	0,  fills.size()+1
    );
    //deviationGraph->GetZaxis()->SetRangeUser( -1, 1 );
    deviationGraphSansBad->GetXaxis()->SetRangeUser(
	0,  neta+1
    );
    fprintf (stderr, "Got here\n" );

    //	Draw the graph.
    gStyle->SetPalette(1);
    deviationGraphSansBad->Draw("surf1");

    deviationGraphSansBad->Write();

    myCanvas->Update();
    ps->NewPage();
    myCanvas->Clear();
    myCanvas->Divide( 1, 1 );
    myCanvas->cd( 1 );

    sprintf ( graphName, "Corrected weighted deviation over time/eta bin" );

    //  Beautifing the graph.
    weightedDeviationGraphSansBad->SetTitle( graphName );

    //	This part of beautifying stuff takes a long time, so we'll warn.
    std::cout 
	<< "Generating graph.  This could take some time. . ." << std::endl;
    weightedDeviationGraphSansBad->GetXaxis()->SetTitle( "\"Fill Time\"" );
    weightedDeviationGraphSansBad->GetYaxis()->SetTitle( "Bin number" );
    weightedDeviationGraphSansBad->GetZaxis()->SetTitle( "Peak Dev. from Mean" );
    weightedDeviationGraphSansBad->GetXaxis()->SetRangeUser(
	0,  fills.size()+1
    );
    //weightedDeviationGraph->GetZaxis()->SetRangeUser( -1, 1 );
    weightedDeviationGraphSansBad->GetXaxis()->SetRangeUser(
	0,  neta+1
    );
    fprintf (stderr, "Got here\n" );

    //	Draw the graph.
    weightedDeviationGraphSansBad->Draw("surf1");

    weightedDeviationGraphSansBad->Write();

    eventsHisto->Write();
    badEventsHisto->Write();

    chi2Histo->Write();
    badChi2Histo->Write();

    outputFile.Close();
    ps->Close();

    std::cout << "Done" << std::endl; 

    return 0;
}

//  This will send an entire html'd table of min/max error points to the stream
//  passed to the function.
void printPointsTable( 
    std::ostream & out,
    DataPoint ** minPoints,
    DataPoint ** maxPoints,
    int neta
)
{
    out << "<h3>Most Significant High/Low Deviation Fills per Eta</h3>"
	<< std::endl;
    out << "<table cellspacing=\"2\" cellpadding=\"2\" border=\"1\">" 
	<< std::endl;
    printPointsHeader( out );
    out << "<tbody>" << std::endl;
    for ( int ii = 0; ii < neta; ++ii )
    {
	printPointRow( out, minPoints[ii], maxPoints[ii] );
    }
    out << "</tbody>" << std::endl;
    out << "</table>" << std::endl;
}

//  This function will just print out an HTML header for the points table to
//  the specified stream.  It may be overkill to have made this a header, but,
//  meh.
void printPointsHeader( std::ostream & out )
{
    out << "<thead>" << std::endl;
    out << "<tr><th rowspan=\"2\">Eta</th><th colspan=\"6\">Min Fill</th><th colspan=\"6\">Max Fill</th></tr>" << std::endl;
    out << "<tr>";
    out << "<th>Fill</th><th>Peak</th><th>Peak Error</th><th>Deviation</th><th>Ratio of Deviation to Mean</th><th>Weighted Deviation</th>";
    out << "<th>Fill</th><th>Peak</th><th>Peak Error</th><th>Deviation</th><th>Ratio of Deviation to Mean</th><th>Weighted Deviation</th>";
    out << "</tr>" << std::endl;
    out << "</thead>" << std::endl;
}

//  This will print an individual row of the min/max error point table.
void printPointRow(
    std::ostream & out, DataPoint * minPoint, DataPoint * maxPoint
)
{
    if ( minPoint->eta != maxPoint->eta )
    {
	fprintf( stderr, "Min/max point eta values did not match!\n" );
	exit(1);
    }
    out << "<tr>" << std::endl;
    out << "<td>" << minPoint->eta << "</td>";
    out << "<td>" << minPoint->fill << "</td>";
    out << "<td>" << minPoint->peak << "</td>";
    out << "<td>" << minPoint->error << "</td>";
    out << "<td>" << minPoint->deviation << "</td>";
    out << "<td>" << minPoint->ratioDeviation << "</td>";
    out << "<td>" << minPoint->weightedDeviation << "</td>";

    out << "<td>" << maxPoint->fill << "</td>";
    out << "<td>" << maxPoint->peak << "</td>";
    out << "<td>" << maxPoint->error << "</td>";
    out << "<td>" << maxPoint->deviation << "</td>";
    out << "<td>" << maxPoint->ratioDeviation << "</td>";
    out << "<td>" << maxPoint->weightedDeviation << "</td>";
    out << "</tr>" << std::endl;
}

//  This will send two HTML tables to output, one for a table of points with
//  very low statistics, and one for a table of points that are known bad.
void printBadPointsTable( 
    std::ostream & out,
    std::set<DataPoint *> lowStat,
    std::set<DataPoint *> bad
)
{
    std::set<DataPoint *>::iterator myIter;
    out << "<h3>Low Statistic Eta-Fills</h3>" << std::endl;
    out << "<table cellspacing=\"2\" cellpadding=\"2\" border=\"1\">" 
	<< std::endl;
    printBadPointsHeader( out );
    out << "<tbody>" << std::endl;
    for ( myIter = lowStat.begin(); myIter != lowStat.end(); ++myIter )
    {
	printBadPointRow( out, *myIter );
    }
    out << "</tbody>" << std::endl;
    out << "</table>" << std::endl;
    out << "<h3>Known Bad Fit Eta-Fills</h3>" << std::endl;
    out << "<table cellspacing=\"2\" cellpadding=\"2\" border=\"1\">" 
	<< std::endl;
    printBadPointsHeader( out );
    out << "<tbody>" << std::endl;
    for ( myIter = bad.begin(); myIter != bad.end(); ++myIter )
    {
	printBadPointRow( out, *myIter );
    }
    out << "</tbody>" << std::endl;
    out << "</table>" << std::endl;
}

//  This function will just print out an HTML header for the bad points table
//  to the specified stream.
void printBadPointsHeader( std::ostream & out )
{
    out << "<thead>" << std::endl;
    out << "<tr>";
    out << "<th>Eta</th><th>Fill</th><th># of Events</th><th>Peak</th><th>Peak Error</th><th>Mean</th><th>Mean Error</th><th>Deviation</th><th>Ratio of Deviation to Mean</th><th>Weighted Deviation</th><th>Chi^2</th><th>DOF</th><th>Chi^2/DOF</th>";
    out << "</tr>" << std::endl;
    out << "</thead>" << std::endl;
}

//  This will print an individual row of the bad point table.
void printBadPointRow( std::ostream & out, DataPoint *point )
{
    out << "<tr>" << std::endl;
    out << "<td>" << point->eta << "</td>";
    out << "<td>" << point->fill << "</td>";
    out << "<td>" << point->events << "</td>";
    out << "<td>" << point->peak << "</td>";
    out << "<td>" << point->error << "</td>";
    out << "<td>" << point->mean << "</td>";
    out << "<td>" << point->meanError << "</td>";
    out << "<td>" << point->deviation << "</td>";
    out << "<td>" << point->ratioDeviation << "</td>";
    out << "<td>" << point->weightedDeviation << "</td>";
    out << "<td>" << point->chi2 << "</td>";
    out << "<td>" << point->ndf << "</td>";
    out << "<td>";
    if ( point->ndf != 0 )
    {
	out << point->chi2ndf;
    }
    out << "</td>";
    out << "</tr>" << std::endl;
}
