#include <iostream>
#include <fstream>
using namespace std;

#include "TFile.h"
#include "TMath.h"
#include "TH1.h"
#include "TF1.h"
#include "TPostScript.h"
#include "TCanvas.h"
#include "TStyle.h"
#include "TLatex.h"
#include "TString.h"
#include "TLine.h"
#include "TRegexp.h"

void drawTower(TH1* h, TF1* f, int id);

//  We're going to assume that there is only one file that we're dealing with
//  here.  This whole "working with multiple sets of histograms" thing seems
//  like overkill at this point.
void fitter(
    const char* file="/star/institutions/uky/wwitzke/mip/0DEE4A066A1446A356054A3970015AA7_0.mip.root",
    //const char* file="/star/institutions/uky/wwitzke/mip/F01007FA2DF760108D7842BB07D3355D_0.mip.root",
    const char* postscript="mip.ps",
    const char* gainfile="mip.gains"
)
{
    gROOT->Macro("$STAR/StRoot/StMuDSTMaker/COMMON/macros/loadSharedLibraries.C");

    //	This is for record keeping in the output files.
    cout << "input file:  " << file << endl;
    cout << "plots:           " << postscript << endl;
    cout << "gains:           " << gainfile << endl;

    //	Setting up the. . ?
    /*
    gStyle->SetCanvasColor( 10 );
    gStyle->SetCanvasBorderMode( 0 );
    gStyle->SetStatColor( 10 );
    gStyle->SetOptTitle( 0 );
    gStyle->SetOptFit( 111 );
    gStyle->SetOptStat( "oume" );
    */

    //	We're going to read file names, line by line, from another file.  It
    //	seems as though there should be a better way to do this (for instance,
    //	wildcards), but, meh, I didn't come up with this scheme.
    TPostScript *ps = new TPostScript( postscript );
    TCanvas *myCanvas = new TCanvas( "c", "", 100, 100, 600., 800. );
    int pad;

    TFile input_file( file, "READ" );

    //  Since we can't know, a priori, the names of the histograms, we're
    //  going to collect all objects from the .root file that we have,
    //  presumeably, opened and try each one.  In order to do this, we read
    //  all first, to make sure that all the file objects (keys?) are read
    //  into memory before we start trying things.  We're also going to get
    //  a count of the number of items in memory and compare afterwards,
    //  just to make sure that this poorly documented development package
    //  isn't doing something funny with memory management here.
    input_file.ReadAll();
    TList *curList = input_file.GetList();
    int numHistograms = curList->GetSize();
    TIter objIter(curList);
    TObject *curObject;

    //print fit stats to a file
    ofstream gains( gainfile );
    char line[500];

    //	Temporary variables used in the loop.
    TString name;
    TString tempString;
    char fitname[200];
    TH1F * tempHisto;
    TF1 * tempFit;
    TLine *gaussian_peak;
    int counter = 0;
    int curEta;

    //	We need to be able to find the eta number from the name of the
    //	histogram with which we are working.  This regular expression should
    //	do that.
    const TRegexp etaFinder("[0-9]+$");

    while ( (curObject = objIter.Next()) )
    {
	//  We need to make sure that the object we're pulling out is some
	//  kind of one-dimensional histogram.  If it's not, then we need to
	//  skip it.
	if ( (tempHisto = dynamic_cast<TH1F *> (curObject) ) == NULL )
	{
	    cout << "Didn't get TH1F" << endl;
	    ++counter;
	    continue;
	}

	if ( tempHisto->Integral( 2.5, 250. ) < 0.5 )
	{
	    cout << "skipped histogram " << name
		 << " (# " << counter + 1 << ")" << endl;

	    ++counter;
	    continue;
	}

	name = tempHisto->GetName();
	tempString = name(etaFinder).Data();
	curEta = tempString.Atoi();

	if ( counter % 15 == 0 )
	{
	    myCanvas->Update();
	    ps->NewPage();
	    myCanvas->Clear();
	    myCanvas->Divide( 3, 5 );
	    pad = 1;
	}

	if ( counter % 600 == 0 )
	    cout << "on histogram " << name
		 << " (# " << counter + 1 << ")" << endl;

	myCanvas->cd( pad );

	tempHisto->Draw("e");

	sprintf ( fitname, "%s_Gaussian_fit", name.Data() );

	tempFit = new TF1( fitname, "landau", 2.5, 250. );
	tempFit->SetParameter( 1, 17. );
	tempFit->SetParameter( 2, 3. );
	//tempFit->SetParLimits( 1, 10, 50 );
	//tempFit->SetParLimits( 2, 0, 25 );

	tempFit->SetLineColor( kYellow );
	tempFit->SetLineWidth( 0.6 );

	tempHisto->Fit( tempFit, "rq" );
	//tempHisto->GetXaxis()->SetRangeUser( 5., 35. );
    
	//double fitMaximum = tempFit->GetMaximum( 0., 100. );
	//double mipPeak = tempFit->GetX( fitMaximum, 0., 100. );
	double mipPeak = tempFit->GetParameter( 1 );
	double mipPeakError = tempFit->GetParError( 1 );

	//  So, we're going to check to see if the mipPeak is less than zero.
	//  If it is, we're going to restrict the range of the fit and see if
	//  that doesn't fix some of the problems that we're seeing.
	if ( mipPeak <= 0 )
	{
	    delete tempFit;
	    tempFit = new TF1( fitname, "landau", 2.5, 80. );
	    tempFit->SetParameter( 1, 17. );
	    tempFit->SetParameter( 2, 3. );

	    tempFit->SetLineColor( kRed );
	    tempFit->SetLineWidth( 0.6 );

	    tempHisto->Fit( tempFit, "rq" );

	    mipPeak = tempFit->GetParameter( 1 );
	    mipPeakError = tempFit->GetParError( 1 );
	}

	drawTower( tempHisto, tempFit, name );
    
	double histogram_top =
	    tempHisto->GetBinContent( tempHisto->GetMaximumBin() );
    
	double gaussian_mean =
	    tempFit->GetParameter( 1 );
	gaussian_peak =
	    new TLine( gaussian_mean, 0., gaussian_mean, histogram_top+15 );
	gaussian_peak->SetLineColor( kGreen );
	gaussian_peak->SetLineWidth( 2.0 );
	gaussian_peak->Draw( "same" );
    
	bool gotError = false;
	/*
	if ( fitMean <= 0 )
	{
	    cout << "fitMean < 0" << endl;
	    gotError = true;
	}
	*/

	if ( mipPeak == 0 )
	{
	    cout << "mipPeak = 0" << endl;
	    gotError = true;
	}

	/*
	if ( mipPeak >= fitMean )
	{
	    cout << "peak >= mean" << endl;
	    gotError = true;
	}
	*/

	if ( gotError )
	{
	    cout << name << endl;
	    cout << "THIS ONE WAS BAD ";
	    cout << "Integral is: " << tempHisto->Integral( 5., 250. ) << endl;
	    //cout << counter << "  " << fitMean << "  " << mipPeak << endl;
	    cout << mipPeak << endl;
	    mipPeak = 0;
	}
	/*
	else
	{
	    cout << "THIS ONE WAS GOOD";
	    cout << "Integral is: " << tempHisto->Integral( 5., 250. ) << endl;
	}
	*/

	float eta = getEta(curEta);
	float theta= 2. * TMath::ATan( TMath::Exp( -eta ) );
	double gain = 0;
	if ( mipPeak != 0 )
	    gain = 0.261*(1.+0.056*eta*eta)/(TMath::Sin(theta)*mipPeak);
	double width = tempFit->GetParameter( 2 );
	
	//sprintf ( line, "%s\t%1.6f\t%2.6f", name.Data(), mipPeak, width );
	//tempHisto->GetXaxis()->SetRangeUser( 5., 50. );
	//sprintf ( line, "%s\t%1.6f\t%2.6f\t%2.6f\t%2.6f", name.Data(), mipPeak, width, tempHisto->GetMean(), tempHisto->GetMeanError() );
	sprintf (
	    line, "%s\t%1.6f\t%2.6f\t%2.6f\t%2.6f\t%i\t%2.6f\t%i",
	    name.Data(),
	    mipPeak,
	    mipPeakError,
	    tempHisto->GetMean(),
	    tempHisto->GetMeanError(),
	    tempHisto->GetEntries(),
	    tempFit->GetChisquare(),
	    tempFit->GetNDF()
	);
	gains << line << endl;

	++pad;
	++counter;

	delete tempFit;
	delete gaussian_peak;
    }
    gains.close();
    ps->Close();

    cout << "saved fit params in " << gainfile << endl;
    if ( numHistograms != counter )
    {
	cout << "Did not get the expected number of histograms" << endl;
    }
}

//  This function will take the pseudco eta bin number associated with the name
//  of the histograms being processed by this fitter and returns the actual
//  eta value.
float getEta( int bin )
{
    float retVal;
    StEmcGeom myEtaFinder("bemc");
    if ( bin > 20 )
    {
	myEtaFinder.getEta( 59, bin - 20, retVal );
    }
    else
    {
	myEtaFinder.getEta( 61, 21 - bin, retVal );
    }

    return retVal;
}

void drawTower( TH1* h, TF1* f, TString name )
{
    //calculate a few quantities
    double peak = f->GetParameter( 1 );
    double mean = f->Mean( 5., 200. );
    double histo_height = h->GetBinContent( h->GetMaximumBin() );
    if ( histo_height == 0 )
	histo_height = 1.;
    
    int xLatexBin = 20;

    //histogram options
    h->SetXTitle( "ADC" );
    h->Draw("e");
    
    //draw a line through the location of the MIP peak
    char line_name[50];
    sprintf ( line_name, "mip_peak_%s", name.Data() );
    TH1* line =
	new TH1F(
	    line_name,
	    line_name,
	    h->GetNbinsX(),
	    h->GetBinLowEdge( 1 ),
	    h->GetBinLowEdge( h->GetNbinsX() + 1 )
	);
    line->SetLineColor( kRed );
    line->Fill( peak, 1.e8 );
//    line->Draw("same");
    
    //write the tower number on the plot
    char tower_title[100];
    sprintf ( tower_title, "%s", name.Data() );//,f->GetParameter(1));
    TLatex title_latex;
    title_latex.SetTextSize( 0.02 );
//    title_latex.DrawLatex(xLatexBin,0.94*histo_height,tower_title);
    title_latex.SetTextAngle( TMath::Pi()/2 );
    title_latex.DrawTextNDC( 0., 0., tower_title );

    delete line;

/*    
    //write the value of the MIP peak
    char tower_peak[100];
    sprintf( tower_peak, "%2.3f", peak );
    TLatex peak_latex;
    peak_latex.SetTextSize( 0.15 );
    peak_latex.DrawLatex( xLatexBin, 0.70 * histo_height, tower_peak );
    
    //write the width of the MIP peak
    char tower_width[100];
    sprintf(tower_width,"%2.3f",f->GetParameter(2));
    TLatex width_latex;
    width_latex.SetTextSize(0.15);
//    width_latex.DrawLatex(xLatexBin,0.46*histo_height,tower_width);
    
    //write the corresponding full-scale energy
    char tower_energy[100];
    double transverse_gain = 0.261*(1.+0.056*eta*eta)/peak;
    sprintf(tower_energy,"%2.3f",4066*transverse_gain);
//    sprintf(tower_energy,"%i:%i:%i",(int)h->GetBinCenter(h->GetMaximumBin()),(int)h->GetBinContent(0),(int)h->GetBinContent(h->GetNbinsX()+1));
//    sprintf(tower_energy,"%2.2f/%i",f->GetChisquare(),f->GetNDF());
    TLatex energy_latex;
    energy_latex.SetTextSize(0.15);
//    energy_latex.DrawLatex(xLatexBin,0.22*histo_height,tower_energy);	
    
    //write if there is a problem with iso or p/E
//    char tower_probs[100];
    TLatex tower_probs;
    tower_probs.SetTextSize(0.25);
    tower_probs.SetTextColor(6);
    
    if(id==3541)
	tower_probs.DrawLatex(1,0.5*histo_height,"begin");
    if(id==4100)
	tower_probs.DrawLatex(1,0.5*histo_height,"end");
*/
    
}
