#include <iostream>
#include <fstream>
#include <cmath>
#include <vector>
#include <map>
#include <algorithm>
#include <stdio.h>
using namespace std;

#include "TPostScript.h"
#include "TCanvas.h"
#include "TStyle.h"
#include "TGraphErrors.h"
#include "TGraph2D.h"

/*  So, the point of this function is to read in a file of fill, eta, peak, and
 *  width values and use that information to construct a graph of the percent
 *  deviations over time for every eta bin.
 */

//  We need a struct here to keep track of all the information that we're
//  reading in from our input file.  We have to use that information in several
//  places, and this seems like the best way to organize that information.
struct DataPoint
{
    int fill = -1;
    int eta = 0;
    double peak = 0;
    double width = 0;
    double mean = 0;
    double meanError = 0;
    //	This is the deviation from the mean of the graph for this point.
    double deviation = 0;
    //	This is the ratio of the devation from the mean for this point to the
    //	mean (i.e. the percent deviation).
    double ratioDeviation = 0;

    DataPoint()
    { }

    //	This is an attempt to solve a memory corruption problem that is
    //	apparently happening with the default copy constructor here.
    DataPoint( const DataPoint & toCopy )
    {
	fprintf(stderr, "Got here: %i, %i\n", toCopy.fill, toCopy.eta);
	fill = toCopy.fill;
	eta = toCopy.fill;
	peak = toCopy.peak;
	width = toCopy.width;
	mean = toCopy.mean;
	meanError = toCopy.meanError;
	deviation = toCopy.deviation;
	ratioDeviation = toCopy.ratioDeviation;
	fprintf( stderr, "Done copying\n" );
    }
};

//  We're going to assume that there is only one file that we're dealing with
//  here.  This whole "working with multiple sets of histograms" thing seems
//  like overkill at this point.
void percentDeviationOverTime(
    const char* file="mip2006.gains.corrected",
    const char* postscript="percentDeviationOverTime.ps"
)
{
    //	This is the number of eta bins that we are expecting.
    const int neta = 40;

    //	This is for record keeping in the output files.
    cout << "input file:  " << file << endl;
    cout << "output file: " << postscript << endl;

    TPostScript *ps = new TPostScript( postscript );
    TCanvas *myCanvas = new TCanvas( "c", "", 100, 100, 600., 800. );

    //	This is our input file.
    ifstream infile( file );

    //	We're going to be reading in all the data from the file into a reusable
    //	data structure (or set of data structures).  I think that, basically,
    //	we want an array of vectors of data points.  That is, we want, one
    //	vector per eta bin, and each vector will be filled with one data point
    //	per fill for that eta bin.
    std::vector<DataPoint> bins[neta];
    DataPoint curPoint;

    for ( int ii = 0; ii < neta; ++ii )
    {
	bins[ii].resize(1000);
    }

    //	We also need to keep track of a set of fills.  Without this, we won't
    //	know whether the individual vectors are missing fills or not.  This set
    //	will be used to determine the "time" axis of the graph that we are
    //	generating.  You may be wondering why this set is a map.  That is
    //	because the fills will be mapped to their particular ordering time
    //	number using this data structure, allowing us to ignore the ordering of
    //	the eta vectors.
    std::map<int, int> fills;

    //	While we have stuff in the file. . . 
    while (
	infile	>> curPoint.fill
		>> curPoint.eta
		>> curPoint.peak
		>> curPoint.width
		>> curPoint.mean
		>> curPoint.meanError
    )
    {
	bins[ curPoint.eta - 1 ].push_back(curPoint);
	//  This next line should create the map element if it does not exist,
	//  and it should have no negative side effects if the element does
	//  exist, except for causing some slow-down, most of which would occur
	//  regardless of how this was done.
	fills[curPoint.fill];
	//curPoint = new DataPoint();
    }

    infile.close();

    //	Now we need to ensure that the ordering of the fills is mapped to the
    //	fill number in the fills map.
    std::map<int, int>::iterator myIter;
    int fillNumber = 0;

    //	This is primarily here because I don't trust libraries implemented by
    //	other people to follow standards.  I also don't trust libraries I've
    //	built myself to follow standards, so this really should not insult
    //	anybody.
    int oldFill = -1;

    for (
	myIter=fills.begin();
	myIter != fills.end();
	++fillNumber, ++myIter
    )
    {
	if (myIter->first < oldFill)
	{
	    cerr << "Iterating through the map did not function as advertised" << endl;
	    return;
	}

	oldFill = myIter->first;
	myIter->second = fillNumber;
    }

    char graphName[200];

    TGraphErrors *myGraph[neta];
    int numPoints;

    //	So, next we create the graphs that we're using to find our fits.  This
    //	should be relatively straight-forward now.
    for ( int ii = 0; ii < neta; ++ii )
    {
	myGraph[ ii ] = new TGraphErrors();
	//  Notice that we're setting the number of points in the graph first.
	myGraph[ ii ]->Set( bins[ii].size() );

	numPoints = bins[ii].size();

	for ( int jj = 0; jj < numPoints; ++jj )
	{
	    myGraph[ ii ]->SetPoint(
		jj, (double) fills[ bins[ii][jj].fill ]+1, bins[ii][jj].peak
	    );

	    //	We're still worried(?) about peak that are equal to zero, so:
	    if ( bins[ii][jj].peak == 0 )
	    {
		cerr << "Got peak == 0 on bin " << ii << ", fill ";
		cerr << bins[ii][jj].fill << endl;
	    }
	}
    }

    //	Now that we have our graphs, we need to fit them.  We expect there to
    //	be a constant-over-time relationship between the points that we are
    //	plotting, so we'll use a constant, linear fit.  We'll need a bunch of
    //	these fitters.
    TF1 *fitters[neta];
    char fitname[200];

    for ( int ii = 0; ii < neta; ++ii )
    {
	sprintf( fitname, "Fit for eta %i", ii+1 );
	//  We're going from 1 to fills size plus 1 here because that should
	//  be the range of the graph (see the graph generation algorithm).
	fitters[ii] = new TF1( fitname, "pol0", 1., fills.size()+1 );

	myGraph[ ii ]->Fit( fitters[ii], "rq" );

	//  Since we're not drawing these graphs, we're done.
    }

    //	Next, we have to construct the 2D graph in time, eta bin, and percent
    //	deviation.  In order to do this, we need to calculate the percent
    //	deviation from the mean, as determined by the previous fit.
    double mean;
    for ( int ii = 0; ii < neta; ++ii )
    {
	mean = fitters[ii]->GetParameter(0);
	//  This is just in case.  Should never happen.
	if ( mean == 0 )
	{
	    cerr << "Got mean = 0 for bin " << ii << ", skipping" << endl;
	    continue;
	}

	numPoints = bins[ii].size();

	for ( int jj = 0; jj < numPoints; ++jj )
	{
	    bins[ii][jj].deviation = bins[ii][jj].peak - mean;
	    bins[ii][jj].ratioDeviation = bins[ii][jj].deviation/mean;
	}
    }

    //	Now we can fill in graph data.  We'll need an independent counter for
    //	the SetPoint function.
    int pointCounter = 0;

    //	And we'll need an actual graph.  We can, however, determine apriori the
    //	maximum number of points that we will be storing in this graph.
    TGraph2D deviationGraph( fills.size() * neta );

    for ( int ii = 0; ii < neta; ++ii )
    {
	numPoints = bins[ii].size();

	for ( int jj = 0; jj < numPoints; ++jj )
	{
	    deviationGraph.SetPoint(
		++pointCounter,
		fills[ bins[ii][jj].fill ] + 1,
		(double) ii + 1,
		bins[ii][jj].percentDeviation
	    );
	}
    }

    myCanvas->cd( 1 );

    sprintf ( graphName, "Percent deviation over time/eta bin" );

    //  Beautifing the graph.
    deviationGraph->SetTitle( graphName );
    deviationGraph->GetXaxis()->SetTitle( "\"Fill Time\"" );
    deviationGraph->GetYaxis()->SetTitle( "Bin number" );
    deviationGraph->GetZaxis()->SetTitle( "Peak Dev. from Mean" );
    //deviationGraph->GetYaxis()->SetRangeUser( 12, 25 );
    deviationGraph->GetXaxis()->SetRangeUser(
	0,  fills.size()
    );
    deviationGraph->GetXaxis()->SetRangeUser(
	0,  neta
    );

    //	Draw the graph.
    gStyle->SetPalette(1);
    deviationGraph->Draw("surf1");

    ps->Close();

    cout << "Done" << endl; 
}
